#include "global.h"
#include "backend.h"

BEPerlVisitor::BEPerlVisitor(Printer *printer)
{
  this->printer = printer;
}

BEPerlVisitor::~BEPerlVisitor()
{
  delete printer;
}

void BEPerlVisitor::visit(ASTConstNumber *peer)
{
  printer->print("%.2f", peer->getValue());
}

void BEPerlVisitor::visit(ASTLetBindingExpression *peer)
{
  printer->print("$%s = ", peer->getNameToBind());
  peer->getRightHandExp()->accept(this);
  printer->println(";");
  peer->getFollowExp()->accept(this);
}

void BEPerlVisitor::visit(ASTTypeDefExpression *peer)
{
  panic("BEPerlVisitor: typedef not implemented");
/*  printer->print("%s = ", peer->getNameToBind());
  peer->getTypeDef()->accept(this);
  printer->println(";");
  peer->getFollowExp()->accept(this); */
}

void BEPerlVisitor::visit(ASTLambdaExpression *peer)
{
  printer->print("sub ");
//  if (peer->getOwnNameOrNull())
//    printer->print("%s ", peer->getOwnNameOrNull());
//  peer->getBindingType()->accept(this);
//  printer->print(") ");
//  if (peer->getReturnTypeOrNull()) {
//    printer->print(": ");
//    peer->getReturnTypeOrNull()->accept(this);
//    printer->print(" ");
//  }
  printer->println("{");
  printer->indent(+1);
  printer->println("my $%s = shift();", peer->getNameToBind());
  peer->getInnerExp()->accept(this);
  printer->println();
  printer->indent(-1);
  printer->print("}");
}

void BEPerlVisitor::visit(ASTBaseType *peer)
{
  panic("BEPerlVisitor: base type not implemented");
//  printer->print("%s", peer->getName());
}

void BEPerlVisitor::visit(ASTInfixOpExpression *peer)
{
  peer->getLeftExpression()->accept(this);
  printer->print(" %c ", peer->getOp());
  peer->getRightExpression()->accept(this);
}

void BEPerlVisitor::visit(ASTVariable *peer)
{
  printer->print("$%s", peer->getName());
}

void BEPerlVisitor::visit(ASTUnitValue *peer)
{
  panic("BEPerlVisitor: unit value not implemented");
//  printer->print("()");
}

void BEPerlVisitor::visit(ASTTypeVariable *peer)
{
  panic("BEPerlVisitor: type variable not implemented");
//  printer->print("%s", peer->getName());
}

void BEPerlVisitor::visit(ASTNamedType *peer)
{
  panic("BEPerlVisitor: named type not implemented");
//  printer->print("%s", peer->getName());
}

void BEPerlVisitor::visit(ASTUnionType *peer)
{
  panic("BEPerlVisitor: union type not implemented");
/*  peer->getOption(0)->accept(this);
  printer->print(" + ");
  peer->getOption(1)->accept(this); */
}

void BEPerlVisitor::visit(ASTTensorType *peer)
{
  panic("BEPerlVisitor: tensor type not implemented");
/*  peer->getLeft()->accept(this);
  printer->print(" * ");
  peer->getRight()->accept(this); */
}

void BEPerlVisitor::visit(ASTAmpersandType *peer)
{
  panic("BEPerlVisitor: ampersand type not implemented");
/*  peer->getLeft()->accept(this);
  printer->print(" & ");
  peer->getRight()->accept(this); */
}

void BEPerlVisitor::visit(ASTFuzzyType *peer)
{
  panic("BEPerlVisitor: fuzzy type not implemented");
/*  printer->print("fuzzy ");
  peer->getInnerType()->accept(this); */
}

void BEPerlVisitor::visit(ASTMuType *peer)
{
  panic("BEPerlVisitor: mu type not implemented");
/*  printer->print("mu %s = ", peer->getName());
  peer->getInnerType()->accept(this); */
}

void BEPerlVisitor::visit(ASTCaseExpression *peer)
{
  panic("BEPerlVisitor: case expression not implemented");
/*  printer->print("case ");
  peer->getExpr()->accept(this);
  printer->println(" {");
  printer->indent(+1);
  printer->println("inl(%s) =>", peer->getId(0));
  printer->indent(+1);
  peer->getBranch(0)->accept(this);
  printer->indent(-1);
  printer->println();
  printer->println("inr(%s) =>", peer->getId(1));
  printer->indent(+1);
  peer->getBranch(1)->accept(this);
  printer->indent(-1);
  printer->println();
  printer->indent(-1);
  printer->print("}"); */
}

void BEPerlVisitor::visit(ASTUnfoldExpression *peer)
{
  panic("BEPerlVisitor: unfold not implemented");
/*  printer->print("unfold ");
  peer->getInnerExp()->accept(this); */
}

void BEPerlVisitor::visit(ASTBuiltinExpression *peer)
{
  panic("BEPerlVisitor: builtin expressions not implemented");
//  printer->print("%s", peer->getName());
}

void BEPerlVisitor::visit(ASTFoldExpression *peer)
{
  panic("BEPerlVisitor: fold not implemented");
/*  printer->print("fold[");
  peer->getTypeArg()->accept(this);
  printer->print("] ");
  peer->getInnerExp()->accept(this); */
}

void BEPerlVisitor::visit(ASTInjectExpression *peer)
{
  panic("BEPerlVisitor: inl/r not implemented");
/*  printer->print("in%c[", peer->getBranch() ? 'r' : 'l');
  peer->getTypeArg()->accept(this);
  printer->print("] ");
  peer->getInnerExp()->accept(this); */
}

void BEPerlVisitor::visit(ASTProjectExpression *peer)
{
  panic("BEPerlVisitor: fst/snd not implemented");
/*  printer->print("%s ", peer->getBranch() ? "snd" : "fst");
  peer->getInnerExp()->accept(this); */
}

void BEPerlVisitor::visit(ASTTensorTuple *peer)
{
  panic("BEPerlVisitor: tensors not implemented");
/*  printer->print("(");
  peer->getLeftExpression()->accept(this);
  printer->print(",");
  peer->getRightExpression()->accept(this);
  printer->print(")"); */
}

void BEPerlVisitor::visit(ASTTensorDestructorExpression *peer)
{
  panic("BEPerlVisitor: tensor destructor not implemented");
/*  printer->print("let (%s,%s) = ", peer->getName(0), peer->getName(1));
  peer->getProducingExp()->accept(this);
  printer->println(";");
  peer->getConsumingExp()->accept(this); */
}

void BEPerlVisitor::visit(ASTAmpersandTuple *peer)
{
  panic("BEPerlVisitor: ampersand tuple not implemented");
/*  printer->print("(|");
  peer->getLeftExpression()->accept(this);
  printer->print(",");
  peer->getRightExpression()->accept(this);
  printer->print("|)"); */
}

void BEPerlVisitor::visit(ASTApplyExpression *peer)
{
  printer->print("(");
  peer->getFun()->accept(this);
  printer->print(")->(");
  peer->getArg()->accept(this);
  printer->print(")");
}

void BEPerlVisitor::visit(ASTUnitType *peer)
{
  panic("BEPerlVisitor: unit type not implemented");
//  printer->print("()");
}

void BEPerlVisitor::visit(ASTFunctionType *peer)
{
  panic("BEPerlVisitor: function type not implemented");
/*  peer->getArgType()->accept(this);
  if (!isinf(peer->getSensitivity()))
    printer->print(" -O[%.2f] ", peer->getSensitivity());
  else
    printer->print(" -> ");

  peer->getResultType()->accept(this); */
}

void BEPerlVisitor::visit(ASTSetType *peer)
{
  panic("BEPerlVisitor: set type not implemented");
/*  peer->getMemberType()->accept(this);
  printer->print(" set"); */
}
