;;; Deprecated runtime for tinyscheme

;;;;; Recursive functions
; (fun fname (x) fbody) = function fname (x : t) { fbody }
(macro (fun form) (let ((fname (cadr form)) 
			(farg (caddr form))
			(fbody (cdddr form))) 
		    `(letrec ((,fname (lambda ,farg ,@fbody)))  ,fname)))

;;;;; Pair types
; (both kinds of pairs are represented as cons cells.)

(define (with-pair x y) (cons x y))
(define (fst x) (car x))
(define (snd x) (cdr x))

(define (tensor-pair x y) (cons x y))

; (tensor-let x y exp body) = let (x,y) = exp in body
(macro (tensor-let form) (let ((v1 (cadr form)) 
			       (v2 (caddr form))
			       (exp (cadddr form))
			       (body (cddddr form))
			       (tmp (gensym))) 
			   `(let* ((,tmp ,exp)
				  (,v1 (car ,tmp)) 
				  (,v2 (cdr ,tmp)))
			      ,@body)))

;;;;; Sum types
(define (inl x) (cons "l" x))
(define (inr x) (cons "r" x))

(macro (sumcase form) (let ((exp (cadr form)) 
			    (b1 (caddr form))
			    (b2 (cadddr form))) 
			`(real-case ,exp 
				    (lambda (,(car b1)) ,@(cdr b1))
				    (lambda (,(car b2)) ,@(cdr b2)))))

(define (real-case e b1 b2) (if (equal? (car e) "l") (b1 (cdr e)) (b2 (cdr e))))

;;;;; Monadic types
; Ot is represented by (maybe effectful) thunks returning t

(define (fuzz x) (lambda () (+ x (lap 1.0))))
(macro (sample form) (let ((var (cadr form)) 
			   (thunk (caddr form))
			   (body (cdddr form))) 
		       `(let ((,var (,thunk))) ,@body)))
(define (return x) (lambda () x))

;;;;; Bags (multisets)
; Represented as lists.

(define (bagsum b) (foldr + 0 b))
(define (bagmap f) (lambda (xs) (map f xs)))
(define (bagsplit f) 
  (define (splitone acc x) (if (f x) 
			       (cons (cons x (car acc)) (cdr acc))
			       (cons (car acc) (cons x (cdr acc)))))
  (lambda (xs) (foldl splitone (cons nil nil) xs)))

;;;;; Clip types

(define (clip x) (if (> x 1) 1 (if (< x 0) 0 x)))


(define (foldl f x lst)
     (if (null? lst)
          x
          (f (foldl f x (cdr lst)) (car lst))))

;;;;; Miscellaneous Things

; alias for equality
(define == equal?)

; alias for unit value
(define nil '())

;;;;; Random numbers
(define rand-max 2147483647)
; uniform random number [0,1]
(define (frand) (/ (rand) rand-max))
; laplace noise with parameter sigma
(define (lap sigma) (let ((s (- (* 2 (modulo (rand) 2)) 1))) (* sigma s (log (frand)))))
; ((random a) b) -> uniform random number in [a,b]
(define (random a) (lambda (b) (+ a (* (- b a) (frand)))))