#include "global.h"
#include "backend.h"

BESchemeVisitor::BESchemeVisitor(Printer *printer)
{
  this->printer = printer;
}

BESchemeVisitor::~BESchemeVisitor()
{
  delete printer;
}

void BESchemeVisitor::visit(ASTConstNumber *peer)
{
  printer->print("%.2f", peer->getValue());
}

void BESchemeVisitor::visit(ASTLetBindingExpression *peer)
{
  printer->print("(let ((%s ", peer->getNameToBind());
  peer->getRightHandExp()->accept(this);
  printer->print("))\n");
  peer->getFollowExp()->accept(this);
  printer->print(")");
}

void BESchemeVisitor::visit(ASTTypeDefExpression *peer)
{
  peer->getFollowExp()->accept(this); 
}

void BESchemeVisitor::visit(ASTLambdaExpression *peer)
{
  const char *name = peer->getOwnNameOrNull();
  if (name) {
    printer->print("(fun ");
    printer->print("%s ", peer->getOwnNameOrNull());
    printer->println("(%s) ", peer->getNameToBind());
  }
  else {
    printer->print("(lambda (%s) ", peer->getNameToBind());
  }
  printer->indent(+1);
  peer->getInnerExp()->accept(this);
  printer->print (")");
  printer->indent(-1);
}

void BESchemeVisitor::visit(ASTBaseType *peer)
{
}

void BESchemeVisitor::visit(ASTInfixOpExpression *peer)
{
  const char *op = peer->getOp();
  if (!strcmp(op, "+")) op = "plus";
  else if (!strcmp(op, "&&")) op = "and";
  else if (!strcmp(op, "||")) op = "or";
  printer->print("(%s ", op);
  peer->getLeftExpression()->accept(this);
  printer->print(" ", peer->getOp());
  peer->getRightExpression()->accept(this);
  printer->print(") ", peer->getOp());
}

void BESchemeVisitor::visit(ASTVariable *peer)
{
  printer->print("%s", peer->getName());
}

void BESchemeVisitor::visit(ASTUnitValue *peer)
{
  printer->print("nil");
}

void BESchemeVisitor::visit(ASTTypeVariable *peer)
{
}

void BESchemeVisitor::visit(ASTNamedType *peer)
{
}

void BESchemeVisitor::visit(ASTUnionType *peer)
{
}

void BESchemeVisitor::visit(ASTTensorType *peer)
{
}

void BESchemeVisitor::visit(ASTAmpersandType *peer)
{
}

void BESchemeVisitor::visit(ASTFuzzyType *peer)
{
}

void BESchemeVisitor::visit(ASTMuType *peer)
{
}

void BESchemeVisitor::visit(ASTCaseExpression *peer)
{
  printer->print("(sumcase ");
  peer->getExpr()->accept(this);
  printer->println();
  printer->indent(+1);
  printer->print("(%s ", peer->getId(0));
  printer->indent(+1);
  peer->getBranch(0)->accept(this);
  printer->indent(-1);
  printer->println(")");
  printer->print("(%s ", peer->getId(1));
  printer->indent(+1);
  peer->getBranch(1)->accept(this);
  printer->indent(-1);
  printer->println(")");
  printer->indent(-1);
  printer->print(")"); 
}

void BESchemeVisitor::visit(ASTUnfoldExpression *peer)
{
  peer->getInnerExp()->accept(this); 
}

void BESchemeVisitor::visit(ASTBuiltinExpression *peer)
{
  printer->print("%s", peer->getName());
}

void BESchemeVisitor::visit(ASTFoldExpression *peer)
{
  peer->getInnerExp()->accept(this); 
}

void BESchemeVisitor::visit(ASTInjectExpression *peer)
{
  printer->print("(in%c ", peer->getBranch() ? 'r' : 'l');
  peer->getInnerExp()->accept(this);
  printer->print(")");
}

void BESchemeVisitor::visit(ASTProjectExpression *peer)
{
  printer->print("(%s", peer->getBranch() ? "snd" : "fst");
  peer->getInnerExp()->accept(this);
  printer->print(")");
}

void BESchemeVisitor::visit(ASTTensorTuple *peer)
{
  printer->print("(tensor-pair ");
  peer->getLeftExpression()->accept(this);
  printer->print(" ");
  peer->getRightExpression()->accept(this);
  printer->print(")");
}

void BESchemeVisitor::visit(ASTTensorDestructorExpression *peer)
{
  printer->print("(tensor-let %s %s ", peer->getName(0), peer->getName(1));
  peer->getProducingExp()->accept(this);
  printer->print(" ");
  peer->getConsumingExp()->accept(this); 
  printer->print(")");
}

void BESchemeVisitor::visit(ASTAmpersandTuple *peer)
{
  printer->print("(with-pair ");
  peer->getLeftExpression()->accept(this);
  printer->print(" ");
  peer->getRightExpression()->accept(this);
  printer->print(")");
}

void BESchemeVisitor::visit(ASTApplyExpression *peer)
{
  printer->print("(");
  peer->getFun()->accept(this);
  printer->print(" ");
  peer->getArg()->accept(this);
  printer->print(")");
}

void BESchemeVisitor::visit(ASTUnitType *peer)
{
}

void BESchemeVisitor::visit(ASTFunctionType *peer)
{
}

void BESchemeVisitor::visit(ASTSetType *peer)
{
}

void BESchemeVisitor::visit(ASTIfThenElseExpression *peer)
{
  printer->print("(if ");
  peer->getExpr()->accept(this);
  printer->print(" ");
  peer->getThenBranch()->accept(this);
  printer->print(" ");
  peer->getElseBranch()->accept(this);
  printer->print(")");
}

void BESchemeVisitor::visit(ASTSampleExpression *peer)
{
  printer->print("(sample %s ", peer->getNameToBind());
  peer->getRightHandExp()->accept(this);
  printer->println(" ");
  peer->getFollowExp()->accept(this);
  printer->print(")");
}

void BESchemeVisitor::visit(ASTConstString *peer)
{
  printer->print("\"");
  printer->print(peer->getValue());
  printer->print("\"");
}
