#include <stdlib.h>
#include <stdio.h>
#include <assert.h>

#include "global.h"
#include "typechecker.h"

TCColons::TCColons()
{
  this->entries = NULL;
}

void TCColons::addSymbol(const char *name, double sensitivity)
{
  struct colonEntry *newEntry = new struct colonEntry;
  newEntry->name = (char*)malloc(strlen(name)+1);
  strcpy(newEntry->name, name);
  newEntry->colon = sensitivity;
  newEntry->next = entries;
  entries = newEntry;
}

struct TCColons::colonEntry *TCColons::findSymbol(const char *name)
{
  struct colonEntry *iter = entries;
  while (iter) {
    if (!strcmp(iter->name, name))
      return iter;
    
    iter = iter->next;
  }
  
  return NULL;
}

void TCColons::setSymbol(const char *name, double sensitivity)
{
  struct colonEntry *entry = findSymbol(name);
  assert(entry);
  
  entry->colon = sensitivity;
}

double TCColons::getSymbol(const char *name)
{
  struct colonEntry *entry = findSymbol(name);
  assert(entry);
  
  return entry->colon;
}

bool TCColons::containsSymbol(const char *name)
{
  return (findSymbol(name) != NULL);
}

void TCColons::removeSymbol(const char *name)
{
  if (!entries)
    return;
    
  if (!strcmp(entries->name, name)) {
    struct colonEntry *iter = entries;
    entries = iter->next;
    free(iter->name);
    free(iter);
    return;
  }
  
  struct colonEntry *iter = entries;
  while (iter->next) {
    if (!strcmp(iter->next->name, name)) {
      struct colonEntry *tmp = iter->next;
      iter->next = tmp->next;
      free(tmp->name);
      free(tmp);
      return;
    }
    
    iter = iter->next;
  }
}

void TCColons::add(TCColons *peer)
{
  struct colonEntry *iter = peer->entries;
  
  while (iter) {
    struct colonEntry *mine = findSymbol(iter->name);
    
    if (mine)
      mine->colon += iter->colon;
    else
      addSymbol(iter->name, iter->colon);
      
    iter = iter->next;
  }
}

void TCColons::takeMax(TCColons *peer)
{
  struct colonEntry *iter = peer->entries;
  
  while (iter) {
    struct colonEntry *mine = findSymbol(iter->name);
    
    if (mine)
      mine->colon = (mine->colon>iter->colon) ? mine->colon : iter->colon;
    else
      addSymbol(iter->name, iter->colon);
      
    iter = iter->next;
  }
}

void TCColons::multiplyBy(double factor)
{
  struct colonEntry *iter = entries;
  while (iter) {
    iter->colon *= factor;
    iter = iter->next;
  }
}

bool TCColons::containsNonzero()
{
  struct colonEntry *iter = entries;
  while (iter) {
    if (iter->colon > 0)
      return true;
      
    iter = iter->next;
  }
  
  return false;
}

char *TCColons::render()
{
  char *buf = (char*)malloc(2000);
  strcpy(buf, "{ ");
  
  struct colonEntry *iter = entries;
  while (iter) {
    strcatf(buf, "%s:%.2f%s ", iter->name, iter->colon, iter->next ? "," : "");
    iter = iter->next;
  }
  
  strcat(buf, "}");
  return buf;
}
