#include <stdarg.h>

#include "global.h"
#include "ast.h"

TCSymbol::TCSymbol(const char *name, ASTType *type, int handle)
{
  if (name) {
    this->name = (char*) malloc(strlen(name)+1);
    strcpy(this->name, name);
  } else {
    this->name = NULL;
  }
  
  this->type = type ? type->clone() : NULL;
  this->prev = NULL;
  this->next = NULL;
  this->handle = handle;
}

TCSymbol::~TCSymbol()
{ 
  assert(prev == NULL);
  assert(next == NULL);
  if (name)
    free(name);
}

char *TCSymbol::render(char *buf, TCSymbol *last, int maxlen, bool first)
{
  if (!first)
    strcat(buf, ", ");
  if (name)
    strcatf(buf, "%s: ", name);
  else
    strcatf(buf, "<unnamed>_%d: ", handle);
    
  if (type) {
    char buf2[2000];
    strcatf(buf, "%s", ::render(type));
  } else {
    strcat(buf, " <unknown>");
  }
    
  if (next != last)
    next->render(buf, last, maxlen, false);
    
  return buf;
}

TCSymbol *TCSymbol::lookup(const char *name, TCSymbol *last)
{
  if (this->name && !strcmp(name, this->name)) 
    return this; 

  if (next != last) 
    return next->lookup(name, last);
    
  return NULL;
}

TCSymbol *TCSymbol::lookupByHandle(int handle, TCSymbol *last)
{
  if (handle == this->handle) 
    return this; 

  if (next != last) 
    return next->lookupByHandle(handle, last);
    
  return NULL;
}

void TCSymbol::enqueueBefore(TCSymbol *sym)
{
  if (!sym) {
    this->next = this;
    this->prev = this;
    return;
  }
    
  this->next = sym;
  this->prev = sym->prev;
  sym->prev->next = this;
  sym->prev = this;
}

void TCSymbol::dequeue()
{
  if (this->next != this) {
    this->next->prev = this->prev;
    this->prev->next = this->next;
  }
   
  prev = NULL;
  next = NULL;
}
