#include "global.h"
#include "ast.h"

#define differenceFound(mine, peer) do { assert(equalSoFar); firstDifference.localSubtree = (mine); firstDifference.peerSubtree = (peer); return; } while (0)
//#define dprintf(a...) printf(a)
#define dprintf(a...) do { } while (0)

class TCReplaceTypeVarVisitor : public ASTDFSVisitor
{
protected:
  ASTTypeVariable *what;
  ASTType *replacement;
  
public:
  TCReplaceTypeVarVisitor(ASTTypeVariable *what, ASTType *replacement);
  virtual void visit(ASTUnionType *myNode);
  virtual void visit(ASTTensorType *myNode);
  virtual void visit(ASTAmpersandType *myNode);
  virtual void visit(ASTFunctionType *myNode);
  virtual void visit(ASTFuzzyType *myNode);
  virtual void visit(ASTSetType *myNode);
};

TCReplaceTypeVarVisitor::TCReplaceTypeVarVisitor(ASTTypeVariable *what, ASTType *replacement) : ASTDFSVisitor()
{
  this->what = what;
  this->replacement = replacement;
}

void TCReplaceTypeVarVisitor::visit(ASTUnionType *peer)
{
  if (peer->getOption(0)->getTypeCode() == TYP_VAR) {
    if (what->equals((ASTTypeVariable*)(peer->getOption(0))))
      peer->setOption(0, replacement->clone());
  }

  if (peer->getOption(1)->getTypeCode() == TYP_VAR) {
    if (what->equals((ASTTypeVariable*)(peer->getOption(1))))
      peer->setOption(1, replacement->clone());
  }
  
  ASTDFSVisitor::visit(peer);
}

void TCReplaceTypeVarVisitor::visit(ASTTensorType *peer)
{
  if (peer->getLeft()->getTypeCode() == TYP_VAR) {
    if (what->equals((ASTTypeVariable*)(peer->getLeft())))
      peer->setLeft(replacement->clone());
  }

  if (peer->getRight()->getTypeCode() == TYP_VAR) {
    if (what->equals((ASTTypeVariable*)(peer->getRight())))
      peer->setRight(replacement->clone());
  }
  
  ASTDFSVisitor::visit(peer);
}

void TCReplaceTypeVarVisitor::visit(ASTAmpersandType *peer)
{
  if (peer->getLeft()->getTypeCode() == TYP_VAR) {
    if (what->equals((ASTTypeVariable*)(peer->getLeft())))
      peer->setLeft(replacement->clone());
  }

  if (peer->getRight()->getTypeCode() == TYP_VAR) {
    if (what->equals((ASTTypeVariable*)(peer->getRight())))
      peer->setRight(replacement->clone());
  }
  
  ASTDFSVisitor::visit(peer);
}

void TCReplaceTypeVarVisitor::visit(ASTFunctionType *peer)
{
  if (peer->getArgType()->getTypeCode() == TYP_VAR) {
    if (what->equals((ASTTypeVariable*)(peer->getArgType())))
      peer->setArgType(replacement->clone());
  }

  if (peer->getResultType()->getTypeCode() == TYP_VAR) {
    if (what->equals((ASTTypeVariable*)(peer->getResultType())))
      peer->setResultType(replacement->clone());
  }
  
  ASTDFSVisitor::visit(peer);
}

void TCReplaceTypeVarVisitor::visit(ASTSetType *peer)
{
  if (peer->getMemberType()->getTypeCode() == TYP_VAR) {
    if (what->equals((ASTTypeVariable*)(peer->getMemberType())))
      peer->setMemberType(replacement->clone());
  }
  
  ASTDFSVisitor::visit(peer);
}

void TCReplaceTypeVarVisitor::visit(ASTFuzzyType *peer)
{
  if (peer->getInnerType()->getTypeCode() == TYP_VAR) {
    if (what->equals((ASTTypeVariable*)(peer->getInnerType())))
      peer->setInnerType(replacement->clone());
  }
  
  ASTDFSVisitor::visit(peer);
}

ASTType *replaceTypeVar(ASTType *where, ASTTypeVariable *what, ASTType *replacement)
{
  assert(where && what && replacement);
  
  char buf1[2000], buf2[2000], buf3[2000];
/*  printf("replaceTypeVar(where=%s, [%s]->[%s])\n",
    render(where, buf1, sizeof(buf1)), render(what, buf2, sizeof(buf2)), render(replacement, buf3, sizeof(buf3))
  ); */
  
  /* Special case if the top-level type is already the variable */
  
  if (where->getTypeCode() == TYP_VAR) {
    if (what->equals((ASTTypeVariable*)where))
      return replacement;
  }
  
  /* Use the visitor */
  
  TCReplaceTypeVarVisitor *visitor = new TCReplaceTypeVarVisitor(what, replacement);
  where->accept(visitor);
  return where;
}

bool ASTType::unify(ASTType **typeA, ASTType **typeB, TCContext *ctx, ASTType **ctxExt)
{ 
  dprintf("--- unify %s|%s\n", render(*typeA), render(*typeB));
  while (true) {
    dprintf("YYY stat %s|%s\n", render(*typeA), render(*typeB));
    ASTType *diffLocal = NULL, *diffPeer = NULL;
    if ((*typeA)->equals(*typeB, ctx, &diffLocal, &diffPeer))
      break;
    
    /* We can only unify if, at the first difference, we have a type variable on at least one side */

    dprintf("ZZZ diff %s|%s\n", render(diffLocal), render(diffPeer));
    assert(diffLocal && diffPeer);    
    if ((diffLocal->getTypeCode() != TYP_VAR) && (diffPeer->getTypeCode() != TYP_VAR))
      return false;
      
    /* OK, this is the case. Swap differences if necessary. */
    
    if (diffLocal->getTypeCode() != TYP_VAR) {
      ASTType *help = diffLocal;
      diffLocal = diffPeer;
      diffPeer = help;
    }
    
    /* Ready to replace */

    dprintf("QQQ Replacing [%s]->[%s]\n", render(diffLocal), render(diffPeer));
    
    assert(diffLocal->getTypeCode() == TYP_VAR);
    *typeA = replaceTypeVar(*typeA, (ASTTypeVariable*)diffLocal, diffPeer);
    *typeB = replaceTypeVar(*typeB, (ASTTypeVariable*)diffLocal, diffPeer);
    if (ctxExt)
      *ctxExt = replaceTypeVar(*ctxExt, (ASTTypeVariable*)diffLocal, diffPeer);
    
    ctx->replaceTypeVarEverywhere((ASTTypeVariable*)diffLocal, diffPeer);
  }
  
  dprintf("--- done\n");
  
  return true;
}
